from ros2run.api import get_executable_path


def launch(launch_descriptor, argv):
    ld = launch_descriptor
    ld.add_process(
        cmd=[get_executable_path(package_name="talker_cpp", executable_name="talker")],
        name="talker",
    )
    ld.add_process(
        cmd=[
            get_executable_path(package_name="listener_py", executable_name="listener")
        ],
        name="listener",
    )

    return ld



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

from testtools.matchers import Equals

from snapcraft.internal import sources
from tests import unit


class TestZip(unit.FakeFileHTTPServerBasedTestCase):
    @mock.patch("zipfile.ZipFile")
    def test_pull_zipfile_must_download_and_extract(self, mock_zip):
        dest_dir = "src"
        os.makedirs(dest_dir)
        zip_file_name = "test.zip"
        source = "http://{}:{}/{file_name}".format(
            *self.server.server_address, file_name=zip_file_name
        )
        zip_source = sources.Zip(source, dest_dir)

        zip_source.pull()

        mock_zip.assert_called_once_with(
            os.path.join(zip_source.source_dir, zip_file_name), "r"
        )

    @mock.patch("zipfile.ZipFile")
    def test_extract_and_keep_zipfile(self, mock_zip):
        zip_file_name = "test.zip"
        source = "http://{}:{}/{file_name}".format(
            *self.server.server_address, file_name=zip_file_name
        )
        dest_dir = os.path.abspath(os.curdir)
        zip_source = sources.Zip(source, dest_dir)

        zip_source.download()
        zip_source.provision(dst=dest_dir, keep_zip=True)

        zip_download = os.path.join(zip_source.source_dir, zip_file_name)
        mock_zip.assert_called_once_with(zip_download, "r")

        with open(zip_download, "r") as zip_file:
            self.assertThat(zip_file.read(), Equals("Test fake file"))

    def test_has_source_handler_entry(self):
        self.assertTrue(sources._source_handler["zip"] is sources.Zip)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
from pathlib import Path
from unittest import mock
from unittest.mock import call

import fixtures
from testtools.matchers import Equals

from snapcraft.internal.repo.apt_cache import AptCache
from tests import unit


class TestAptStageCache(unit.TestCase):
    # This are expensive tests, but is much more valuable than using mocks.
    # When adding tests, consider adding it to test_stage_packages(), or
    # create mocks.
    @unittest.skip("hangs on google spread test with 'Error in function start'")
    def test_stage_packages(self):
        fetch_dir_path = Path(self.path, "debs")
        fetch_dir_path.mkdir(exist_ok=True, parents=True)
        stage_cache = Path(self.path, "cache")
        stage_cache.mkdir(exist_ok=True, parents=True)

        with AptCache(stage_cache=stage_cache) as apt_cache:
            apt_cache.update()

            package_names = {"pciutils"}
            filtered_names = {"base-files", "libc6", "libkmod2", "libudev1", "zlib1g"}

            apt_cache.mark_packages(package_names)
            apt_cache.unmark_packages(unmark_names=filtered_names)

            marked_packages = apt_cache.get_packages_marked_for_installation()
            self.assertThat(
                sorted([name for name, _ in marked_packages]),
                Equals(["libpci3", "pciutils"]),
            )

            names = []
            for pkg_name, pkg_version, dl_path in apt_cache.fetch_archives(
                fetch_dir_path
            ):
                names.append(pkg_name)
                self.assertThat(dl_path.exists(), Equals(True))
                self.assertThat(dl_path.parent, Equals(fetch_dir_path))
                self.assertThat(isinstance(pkg_version, str), Equals(True))

            self.assertThat(sorted(names), Equals(["libpci3", "pciutils"]))


class TestMockedApt(unit.TestCase):
    def test_stage_cache(self):
        stage_cache = Path(self.path, "cache")
        stage_cache.mkdir(exist_ok=True, parents=True)
        self.fake_apt = self.useFixture(
            fixtures.MockPatch("snapcraft.internal.repo.apt_cache.apt")
        ).mock

        with AptCache(stage_cache=stage_cache) as apt_cache:
            apt_cache.update()

        self.assertThat(
            self.fake_apt.mock_calls,
            Equals(
                [
                    call.apt_pkg.config.set("Apt::Install-Recommends", "False"),
                    call.apt_pkg.config.set(
                        "Acquire::AllowInsecureRepositories", "False"
                    ),
                    call.apt_pkg.config.set(
                        "Dir::Etc::Trusted", "/etc/apt/trusted.gpg"
                    ),
                    call.apt_pkg.config.set(
                        "Dir::Etc::TrustedParts", "/etc/apt/trusted.gpg.d/"
                    ),
                    call.apt_pkg.config.clear("APT::Update::Post-Invoke-Success"),
                    call.progress.text.AcquireProgress(),
                    call.Cache(memonly=True, rootdir=str(stage_cache)),
                    call.Cache().update(fetch_progress=mock.ANY, sources_list=None),
                    call.Cache().close(),
                    call.Cache(memonly=True, rootdir=str(stage_cache)),
                    call.Cache().close(),
                ]
            ),
        )

    def test_stage_cache_in_snap(self):
        self.fake_apt = self.useFixture(
            fixtures.MockPatch("snapcraft.internal.repo.apt_cache.apt")
        ).mock

        stage_cache = Path(self.path, "cache")
        stage_cache.mkdir(exist_ok=True, parents=True)

        snap = Path(self.path, "snap")
        snap.mkdir(exist_ok=True, parents=True)

        self.useFixture(
            fixtures.MockPatch("snapcraft.internal.common.is_snap", return_value=True)
        )
        self.useFixture(fixtures.EnvironmentVariable("SNAP", str(snap)))

        with AptCache(stage_cache=stage_cache) as apt_cache:
            apt_cache.update()

        self.assertThat(
            self.fake_apt.mock_calls,
            Equals(
                [
                    call.apt_pkg.config.set("Apt::Install-Recommends", "False"),
                    call.apt_pkg.config.set(
                        "Acquire::AllowInsecureRepositories", "False"
                    ),
                    call.apt_pkg.config.set("Dir", str(Path(snap, "usr/lib/apt"))),
                    call.apt_pkg.config.set(
                        "Dir::Bin::methods",
                        str(Path(snap, "usr/lib/apt/methods")) + os.sep,
                    ),
                    call.apt_pkg.config.set(
                        "Dir::Bin::solvers::",
                        str(Path(snap, "usr/lib/apt/solvers")) + os.sep,
                    ),
                    call.apt_pkg.config.set(
                        "Dir::Bin::apt-key", str(Path(snap, "usr/bin/apt-key"))
                    ),
                    call.apt_pkg.config.set(
                        "Apt::Key::gpgvcommand", str(Path(snap, "usr/bin/gpgv"))
                    ),
                    call.apt_pkg.config.set(
                        "Dir::Etc::Trusted", "/etc/apt/trusted.gpg"
                    ),
                    call.apt_pkg.config.set(
                        "Dir::Etc::TrustedParts", "/etc/apt/trusted.gpg.d/"
                    ),
                    call.apt_pkg.config.clear("APT::Update::Post-Invoke-Success"),
                    call.progress.text.AcquireProgress(),
                    call.Cache(memonly=True, rootdir=str(stage_cache)),
                    call.Cache().update(fetch_progress=mock.ANY, sources_list=None),
                    call.Cache().close(),
                    call.Cache(memonly=True, rootdir=str(stage_cache)),
                    call.Cache().close(),
                ]
            ),
        )

    def test_host_cache_setup(self):
        self.fake_apt = self.useFixture(
            fixtures.MockPatch("snapcraft.internal.repo.apt_cache.apt")
        ).mock

        with AptCache() as _:
            pass

        self.assertThat(
            self.fake_apt.mock_calls, Equals([call.Cache(), call.Cache().close()])
        )


class TestAptReadonlyHostCache(unit.TestCase):
    def test_host_is_package_valid(self):
        with AptCache() as apt_cache:
            self.assertThat(apt_cache.is_package_valid("apt"), Equals(True))
            self.assertThat(
                apt_cache.is_package_valid("fake-news-bears"), Equals(False)
            )

    def test_host_get_installed_packages(self):
        with AptCache() as apt_cache:
            installed_packages = apt_cache.get_installed_packages()
            self.assertThat(isinstance(installed_packages, dict), Equals(True))
            self.assertThat("apt" in installed_packages, Equals(True))
            self.assertThat("fake-news-bears" in installed_packages, Equals(False))

    def test_host_get_installed_version(self):
        with AptCache() as apt_cache:
            self.assertThat(
                isinstance(apt_cache.get_installed_version("apt"), str), Equals(True)
            )
            self.assertThat(
                apt_cache.get_installed_version("fake-news-bears"), Equals(None)
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import testscenarios
from testtools.matchers import FileExists

from tests import integration


class LocalSourceTestCase(integration.TestCase):
    def test_build_local_source(self):
        self.run_snapcraft("build", "local-source")

        self.assertThat(
            os.path.join(self.parts_dir, "make-project", "build", "stamp-all"),
            FileExists(),
        )

    def test_stage_local_source(self):
        self.run_snapcraft("stage", "local-source")

        self.assertThat(
            os.path.join(self.parts_dir, "make-project", "build", "stamp-install"),
            FileExists(),
        )


class LocalSourceTypeTestCase(integration.TestCase):
    def test_build_local_source(self):
        self.run_snapcraft("build", "local-source-type")

        self.assertThat(
            os.path.join(self.parts_dir, "make-project", "build", "stamp-all"),
            FileExists(),
        )

    def test_stage_local_source(self):
        self.run_snapcraft("stage", "local-source")

        self.assertThat(
            os.path.join(self.parts_dir, "make-project", "build", "stamp-install"),
            FileExists(),
        )


class LocalSourceSubfoldersTestCase(testscenarios.WithScenarios, integration.TestCase):

    scenarios = [
        ("Top folder", {"subfolder": "."}),
        ("Sub folder Level 1", {"subfolder": "packaging"}),
        (
            "Sub folder Level 2",
            {"subfolder": os.path.join("packaging", "snap-package")},
        ),
        (
            "Sub folder Level 3",
            {"subfolder": os.path.join("packaging", "snap-package", "yes-really-deep")},
        ),
    ]

    def test_pull_local_source(self):
        self.copy_project_to_cwd("local-source-subfolders")
        os.chdir(self.subfolder)
        self.run_snapcraft("pull")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.repo import errors


class TestErrorFormatting:

    scenarios = (
        (
            "SnapdConnectionError",
            {
                "exception_class": errors.SnapdConnectionError,
                "kwargs": {"snap_name": "test", "url": "url"},
                "expected_message": (
                    "Failed to get information for snap 'test': "
                    "could not connect to 'url'."
                ),
            },
        ),
    )

    def test_error_formatting(self, exception_class, kwargs, expected_message):
        assert str(exception_class(**kwargs)) == expected_message


class TestAptGPGKeyInstallError:

    scenarios = [
        (
            "AptGPGKeyInstallError basic",
            {
                "exception_class": errors.AptGPGKeyInstallError,
                "kwargs": {"output": "some error", "key": "fake key"},
                "expected_brief": "Failed to install GPG key: some error",
                "expected_resolution": "Verify any configured GPG keys.",
                "expected_details": "GPG key:\nfake key\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "AptGPGKeyInstallError basic with warning",
            {
                "exception_class": errors.AptGPGKeyInstallError,
                "kwargs": {
                    "output": "Warning: apt-key output should not be parsed (stdout is not a terminal)\nsome error",
                    "key": "fake key",
                },
                "expected_brief": "Failed to install GPG key: some error",
                "expected_resolution": "Verify any configured GPG keys.",
                "expected_details": "GPG key:\nfake key\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "AptGPGKeyInstallError basic with warning",
            {
                "exception_class": errors.AptGPGKeyInstallError,
                "kwargs": {
                    "output": "Warning: apt-key output should not be parsed (stdout is not a terminal)\nsome error",
                    "key": "fake key",
                },
                "expected_brief": "Failed to install GPG key: some error",
                "expected_resolution": "Verify any configured GPG keys.",
                "expected_details": "GPG key:\nfake key\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "AptGPGKeyInstallError keyserver no data",
            {
                "exception_class": errors.AptGPGKeyInstallError,
                "kwargs": {
                    "output": "gpg: keyserver receive failed: No data",
                    "key_id": "fake-key-id",
                    "key_server": "fake.key.server",
                },
                "expected_brief": "Failed to install GPG key: GPG key 'fake-key-id' not found on key server 'fake.key.server'",
                "expected_resolution": "Verify any configured GPG keys.",
                "expected_details": "GPG key ID: fake-key-id\nGPG key server: fake.key.server\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "AptGPGKeyInstallError keyserver failure",
            {
                "exception_class": errors.AptGPGKeyInstallError,
                "kwargs": {
                    "output": "gpg: keyserver receive failed: Server indicated a failure",
                    "key_id": "fake-key-id",
                    "key_server": "fake.key.server",
                },
                "expected_brief": "Failed to install GPG key: unable to establish connection to key server 'fake.key.server'",
                "expected_resolution": "Verify any configured GPG keys.",
                "expected_details": "GPG key ID: fake-key-id\nGPG key server: fake.key.server\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "AptGPGKeyInstallError keyserver timeout",
            {
                "exception_class": errors.AptGPGKeyInstallError,
                "kwargs": {
                    "output": "gpg: keyserver receive failed: Connection timed out",
                    "key_id": "fake-key-id",
                    "key_server": "fake.key.server",
                },
                "expected_brief": "Failed to install GPG key: unable to establish connection to key server 'fake.key.server' (connection timed out)",
                "expected_resolution": "Verify any configured GPG keys.",
                "expected_details": "GPG key ID: fake-key-id\nGPG key server: fake.key.server\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
    ]

    def test_snapcraft_exception_handling(
        self,
        exception_class,
        expected_brief,
        expected_details,
        expected_docs_url,
        expected_reportable,
        expected_resolution,
        kwargs,
    ):
        exception = exception_class(**kwargs)

        assert exception.get_brief() == expected_brief
        assert exception.get_resolution() == expected_resolution
        assert exception.get_details() == expected_details
        assert exception.get_docs_url() == expected_docs_url
        assert exception.get_reportable() == expected_reportable


def test_packages_not_found_error():
    exception = errors.PackagesNotFoundError(["foo"])

    assert (
        exception.get_brief()
        == "Failed to find installation candidate for packages: 'foo'"
    )
    assert (
        exception.get_resolution()
        == "Verify APT repository configuration and package names are correct."
    )
    assert exception.get_details() is None
    assert exception.get_docs_url() is None
    assert exception.get_reportable() is False


def test_multiple_packages_not_found_error():
    exception = errors.PackagesNotFoundError(["foo", "foo2", "foo3"])

    assert (
        exception.get_brief()
        == "Failed to find installation candidate for packages: 'foo', 'foo2', and 'foo3'"
    )
    assert (
        exception.get_resolution()
        == "Verify APT repository configuration and package names are correct."
    )
    assert exception.get_details() is None
    assert exception.get_docs_url() is None
    assert exception.get_reportable() is False



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
from typing import Dict, Optional  # noqa: F401

_ID_TO_UBUNTU_CODENAME = {
    "17.10": "artful",
    "17.04": "zesty",
    "16.04": "xenial",
    "14.04": "trusty",
}


def get_version_codename() -> Optional[str]:
    """Return the OS version codename

    This first tries to use the VERSION_CODENAME. If that's missing, it
    tries to use the VERSION_ID to figure out the codename on its own.

    """
    os_release = {}  # type: Dict[str, str]
    with contextlib.suppress(FileNotFoundError):
        with open("/etc/os-release") as f:
            for line in f:
                entry = line.rstrip().split("=")
                if len(entry) == 2:
                    os_release[entry[0]] = entry[1].strip('"')

    with contextlib.suppress(KeyError):
        return os_release["VERSION_CODENAME"]

    with contextlib.suppress(KeyError):
        return _ID_TO_UBUNTU_CODENAME[os_release["VERSION_ID"]]

    return None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016, 2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The WAF plugin is useful to build waf based parts

waf bases projects are projects that drive configuration and build via
a local waf python helper - see https://github.com/waf-project/waf for more
details.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

In addition, this plugin uses the following plugin-specific keywords:

    - configflags:
      (list of strings)
      configure flags to pass to the build such as those shown by running
      ./waf --help
"""

from snapcraft.plugins.v1 import PluginV1


class WafPlugin(PluginV1):
    """plugin to build via waf build system"""

    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["configflags"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["required"] = ["source"]

        return schema

    def __init__(self, name, options, project):
        super().__init__(name, options, project)
        self._setup_base_tools()

    def _setup_base_tools(self):
        self.build_packages.append("python-dev:native")

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["configflags"]

    def env(self, root):
        env = super().env(root)
        if self.project.is_cross_compiling:
            env.extend(
                [
                    "CC={}-gcc".format(self.project.arch_triplet),
                    "CXX={}-g++".format(self.project.arch_triplet),
                ]
            )
        return env

    def enable_cross_compilation(self):
        # Let snapcraft know that this plugin can cross-compile
        # If the method isn't implemented an exception is raised
        pass

    def build(self):
        super().build()
        self.run(["./waf", "distclean"])
        self.run(["./waf", "configure"] + self.options.configflags)
        self.run(["./waf", "build"])
        self.run(
            ["./waf", "install", "--destdir=" + self.installdir]
        )  # target from snappy env



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import socketserver
import tempfile
import threading
from unittest import mock

import fixtures

from tests.fake_servers import snapd


class UnixHTTPServer(socketserver.UnixStreamServer):
    def get_request(self):
        request, client_address = self.socket.accept()
        # BaseHTTPRequestHandler expects a tuple with the client address at
        # index 0, so we fake one
        if len(client_address) == 0:
            client_address = (self.server_address,)
        return (request, client_address)


class FakeSnapd(fixtures.Fixture):
    @property
    def snaps_result(self):
        self.request_handler.snaps_result

    @snaps_result.setter
    def snaps_result(self, value):
        self.request_handler.snaps_result = value

    @property
    def snap_details_func(self):
        self.request_handler.snap_details_func

    @snap_details_func.setter
    def snap_details_func(self, value):
        self.request_handler.snap_details_func = value

    @property
    def find_result(self):
        self.request_handler.find_result

    @find_result.setter
    def find_result(self, value):
        self.request_handler.find_result = value

    def __init__(self):
        super().__init__()
        self.request_handler = snapd.FakeSnapdRequestHandler
        self.snaps_result = []
        self.find_result = []
        self.snap_details_func = None

    def setUp(self):
        super().setUp()
        snapd_fake_socket_path = tempfile.mkstemp()[1]
        os.unlink(snapd_fake_socket_path)

        socket_path_patcher = mock.patch(
            "snapcraft.internal.repo.snaps.get_snapd_socket_path_template"
        )
        mock_socket_path = socket_path_patcher.start()
        mock_socket_path.return_value = "http+unix://{}/v2/{{}}".format(
            snapd_fake_socket_path.replace("/", "%2F")
        )
        self.addCleanup(socket_path_patcher.stop)

        self._start_fake_server(snapd_fake_socket_path)

    def _start_fake_server(self, socket):
        self.server = UnixHTTPServer(socket, self.request_handler)
        server_thread = threading.Thread(target=self.server.serve_forever)
        server_thread.start()
        self.addCleanup(self._stop_fake_server, server_thread)

    def _stop_fake_server(self, thread):
        self.server.shutdown()
        self.server.socket.close()
        thread.join()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.build_providers import errors


class TestErrorFormatting:

    scenarios = [
        (
            "ProviderNotSupportedError",
            dict(
                exception_class=errors.ProviderNotSupportedError,
                kwargs=dict(provider="docker"),
                expected_message=(
                    "The 'docker' provider is not supported, please choose a "
                    "different one and try again."
                ),
            ),
        ),
        (
            "ProviderNotFound",
            dict(
                exception_class=errors.ProviderNotFound,
                kwargs=dict(
                    provider="multipass",
                    prompt_installable=False,
                    error_message="not installed",
                ),
                expected_message=(
                    "You need 'multipass' set-up to build snaps: not installed."
                ),
            ),
        ),
        (
            "ProviderLaunchError (exit code)",
            dict(
                exception_class=errors.ProviderLaunchError,
                kwargs=dict(provider_name="multipass", exit_code=1),
                expected_message=(
                    "An error occurred with the instance when trying to launch with "
                    "'multipass': returned exit code 1.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderLaunchError (error message)",
            dict(
                exception_class=errors.ProviderLaunchError,
                kwargs=dict(
                    provider_name="multipass", error_message="failed to launch"
                ),
                expected_message=(
                    "An error occurred with the instance when trying to launch with "
                    "'multipass': failed to launch.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderLaunchError (exit code and error message)",
            dict(
                exception_class=errors.ProviderLaunchError,
                kwargs=dict(
                    provider_name="multipass",
                    exit_code=1,
                    error_message="failed to launch",
                ),
                expected_message=(
                    "An error occurred with the instance when trying to launch with "
                    "'multipass': returned exit code 1: failed to launch.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderStopError (exit code)",
            dict(
                exception_class=errors.ProviderStopError,
                kwargs=dict(provider_name="multipass", exit_code=1),
                expected_message=(
                    "An error occurred with the instance when trying to stop with "
                    "'multipass': returned exit code 1.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderStopError (error message)",
            dict(
                exception_class=errors.ProviderStopError,
                kwargs=dict(provider_name="multipass", error_message="failed to stop"),
                expected_message=(
                    "An error occurred with the instance when trying to stop with "
                    "'multipass': failed to stop.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderStopError (exit code and error message)",
            dict(
                exception_class=errors.ProviderStopError,
                kwargs=dict(
                    provider_name="multipass",
                    exit_code=1,
                    error_message="failed to stop",
                ),
                expected_message=(
                    "An error occurred with the instance when trying to stop with "
                    "'multipass': returned exit code 1: failed to stop.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderDeleteError (exit code)",
            dict(
                exception_class=errors.ProviderDeleteError,
                kwargs=dict(provider_name="multipass", exit_code=1),
                expected_message=(
                    "An error occurred with the instance when trying to delete with "
                    "'multipass': returned exit code 1.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderDeleteError (error message)",
            dict(
                exception_class=errors.ProviderDeleteError,
                kwargs=dict(
                    provider_name="multipass", error_message="failed to delete"
                ),
                expected_message=(
                    "An error occurred with the instance when trying to delete with "
                    "'multipass': failed to delete.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderDeleteError (exit code and error message)",
            dict(
                exception_class=errors.ProviderDeleteError,
                kwargs=dict(
                    provider_name="multipass",
                    exit_code=1,
                    error_message="failed to delete",
                ),
                expected_message=(
                    "An error occurred with the instance when trying to delete with "
                    "'multipass': returned exit code 1: failed to delete.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderShellError (exit code)",
            dict(
                exception_class=errors.ProviderShellError,
                kwargs=dict(provider_name="multipass", exit_code=1),
                expected_message=(
                    "An error occurred with the instance when trying to shell with "
                    "'multipass': returned exit code 1.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderShellError (error message)",
            dict(
                exception_class=errors.ProviderShellError,
                kwargs=dict(provider_name="multipass", error_message="failed to shell"),
                expected_message=(
                    "An error occurred with the instance when trying to shell with "
                    "'multipass': failed to shell.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderShellError (exit code and error message)",
            dict(
                exception_class=errors.ProviderShellError,
                kwargs=dict(
                    provider_name="multipass",
                    exit_code=1,
                    error_message="failed to shell",
                ),
                expected_message=(
                    "An error occurred with the instance when trying to shell with "
                    "'multipass': returned exit code 1: failed to shell.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderMountError (exit code)",
            dict(
                exception_class=errors.ProviderMountError,
                kwargs=dict(provider_name="multipass", exit_code=1),
                expected_message=(
                    "An error occurred with the instance when trying to mount with "
                    "'multipass': returned exit code 1.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderMountError (error message)",
            dict(
                exception_class=errors.ProviderMountError,
                kwargs=dict(provider_name="multipass", error_message="failed to mount"),
                expected_message=(
                    "An error occurred with the instance when trying to mount with "
                    "'multipass': failed to mount.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderMountError (exit code and error message)",
            dict(
                exception_class=errors.ProviderMountError,
                kwargs=dict(
                    provider_name="multipass",
                    exit_code=1,
                    error_message="failed to mount",
                ),
                expected_message=(
                    "An error occurred with the instance when trying to mount with "
                    "'multipass': returned exit code 1: failed to mount.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderFileCopyError (exit code)",
            dict(
                exception_class=errors.ProviderFileCopyError,
                kwargs=dict(provider_name="multipass", exit_code=1),
                expected_message=(
                    "An error occurred with the instance when trying to copy files with "
                    "'multipass': returned exit code 1.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderFileCopyError (error message)",
            dict(
                exception_class=errors.ProviderFileCopyError,
                kwargs=dict(
                    provider_name="multipass", error_message="failed to copy files"
                ),
                expected_message=(
                    "An error occurred with the instance when trying to copy files with "
                    "'multipass': failed to copy files.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderFileCopyError (exit code and error message)",
            dict(
                exception_class=errors.ProviderFileCopyError,
                kwargs=dict(
                    provider_name="multipass",
                    exit_code=1,
                    error_message="failed to copy files",
                ),
                expected_message=(
                    "An error occurred with the instance when trying to copy files with "
                    "'multipass': returned exit code 1: failed to copy files.\n"
                    "Ensure that 'multipass' is setup correctly and try again."
                ),
            ),
        ),
        (
            "ProviderExecError",
            dict(
                exception_class=errors.ProviderExecError,
                kwargs=dict(
                    provider_name="multipass",
                    command=["snap", "install", "snapcraft", "--classic"],
                    exit_code=1,
                ),
                expected_message=(
                    "An error occurred when trying to execute "
                    "'snap install snapcraft --classic' with 'multipass': "
                    "returned exit code 1."
                ),
            ),
        ),
        (
            "ProviderInfoError",
            dict(
                exception_class=errors.ProviderInfoError,
                kwargs=dict(provider_name="multipass", exit_code=1, stderr=b"error"),
                expected_message=(
                    "An error occurred when using 'multipass' to query the status "
                    "of the instance: returned exit code 1: error."
                ),
            ),
        ),
        (
            "ProviderInstanceNotFoundError",
            dict(
                exception_class=errors.ProviderInstanceNotFoundError,
                kwargs=dict(instance_name="test-build"),
                expected_message="Cannot find an instance named 'test-build'.",
            ),
        ),
        (
            "ProviderInfoDataKeyError",
            dict(
                exception_class=errors.ProviderInfoDataKeyError,
                kwargs=dict(
                    provider_name="multipass",
                    missing_key="instance-name",
                    data=dict(info=dict(another_instance="data")),
                ),
                expected_message=(
                    "The data returned by 'multipass' was not expected. It is "
                    "missing a required key "
                    "'instance-name' in {'info': {'another_instance': 'data'}}."
                ),
            ),
        ),
        (
            "ProviderBadDataError",
            dict(
                exception_class=errors.ProviderBadDataError,
                kwargs=dict(provider_name="multipass", data="bad-json"),
                expected_message=(
                    "The data returned by 'multipass' was not expected or in the "
                    "wrong format: 'bad-json'."
                ),
            ),
        ),
        (
            "ProviderCommunicationError",
            dict(
                exception_class=errors.ProviderCommunicationError,
                kwargs=dict(provider_name="multipass", message="failed to foo"),
                expected_message=(
                    "An error occurred when trying to communicate with the "
                    "'multipass' provider: failed to foo."
                ),
            ),
        ),
        (
            "ProviderMultipassDownloadFailed",
            dict(
                exception_class=errors.ProviderMultipassDownloadFailed,
                kwargs=dict(message="failed to foo"),
                expected_message=(
                    "Failed to download Multipass: 'failed to foo'\n"
                    "Please install manually. You can find the latest release at:\n"
                    "https://multipass.run"
                ),
            ),
        ),
        (
            "ProviderMultipassInstallationFailed",
            dict(
                exception_class=errors.ProviderMultipassInstallationFailed,
                kwargs=dict(message="failed to foo"),
                expected_message=(
                    "Failed to install Multipass: 'failed to foo'\n"
                    "Please install manually. You can find the latest release at:\n"
                    "https://multipass.run"
                ),
            ),
        ),
    ]

    def test_error_formatting(self, exception_class, kwargs, expected_message):
        assert str(exception_class(**kwargs)) == expected_message


class TestSnapcraftExceptionTests:
    scenarios = (
        (
            "ProviderInvalidBaseError",
            {
                "exception_class": errors.ProviderInvalidBaseError,
                "kwargs": {"provider_name": "LXD", "build_base": "core29"},
                "expected_brief": "The 'LXD' provider does not support base 'core29'",
                "expected_resolution": "Ensure build-base or base are correct in the snapcraft.yaml file.",
                "expected_details": None,
                "expected_docs_url": "https://snapcraft.io/docs/base-snaps",
                "expected_reportable": False,
            },
        ),
    )

    def test_snapcraft_exception_handling(
        self,
        exception_class,
        expected_brief,
        expected_details,
        expected_docs_url,
        expected_reportable,
        expected_resolution,
        kwargs,
    ):
        exception = exception_class(**kwargs)

        assert exception.get_brief() == expected_brief
        assert exception.get_resolution() == expected_resolution
        assert exception.get_details() == expected_details
        assert exception.get_docs_url() == expected_docs_url
        assert exception.get_reportable() == expected_reportable



import os

from setuptools import find_packages, setup

package_name = "listener_py"

setup(
    name=package_name,
    version="0.0.0",
    packages=find_packages(),
    data_files=[
        (os.path.join("share", package_name), ["package.xml"]),
        (os.path.join("share", package_name, "launch"), ["launch/talk_and_listen.py"]),
    ],
    install_requires=["setuptools"],
    entry_points={"console_scripts": ["listener = nodes.listener:main"]},
)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import subprocess
import sys
from typing import List

from . import errors
from ._base import Base


class Git(Base):
    @classmethod
    def version(self):
        """Get git version information."""
        return (
            subprocess.check_output(["git", "version"], stderr=subprocess.DEVNULL)
            .decode(sys.getfilesystemencoding())
            .strip()
        )

    @classmethod
    def check_command_installed(cls) -> bool:
        """Check if git is installed."""
        try:
            cls.version()
        except FileNotFoundError:
            return False
        return True

    @classmethod
    def generate_version(cls, *, source_dir=None):
        """Return the latest git tag from PWD or defined source_dir.

        The output depends on the use of annotated tags and will return
        something like: '2.28+git.10.abcdef' where '2.28 is the
        tag, '+git' indicates there are commits ahead of the tag, in
        this case it is '10' and the latest commit hash begins with
        'abcdef'. If there are no tags or the revision cannot be
        determined, this will return 0 as the tag and only the commit
        hash of the latest commit.
        """
        if not source_dir:
            source_dir = os.getcwd()

        encoding = sys.getfilesystemencoding()
        try:
            output = (
                subprocess.check_output(
                    ["git", "-C", source_dir, "describe", "--dirty"],
                    stderr=subprocess.DEVNULL,
                )
                .decode(encoding)
                .strip()
            )
        except subprocess.CalledProcessError:
            # If we fall into this exception it is because the repo is not
            # tagged at all.
            proc = subprocess.Popen(
                ["git", "-C", source_dir, "describe", "--dirty", "--always"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            stdout, stderr = proc.communicate()
            if proc.returncode != 0:
                # This most likely means the project we are in is not driven
                # by git.
                raise errors.VCSError(message=stderr.decode(encoding).strip())
            return "0+git.{}".format(stdout.decode(encoding).strip())

        m = re.search(
            r"^(?P<tag>[a-zA-Z0-9.+~-]+)-"
            r"(?P<revs_ahead>\d+)-"
            r"g(?P<commit>[0-9a-fA-F]+(?:-dirty)?)$",
            output,
        )

        if not m:
            # This means we have a pure tag
            return output

        tag = m.group("tag")
        revs_ahead = m.group("revs_ahead")
        commit = m.group("commit")

        return "{}+git{}.{}".format(tag, revs_ahead, commit)

    def __init__(
        self,
        source,
        source_dir,
        source_tag=None,
        source_commit=None,
        source_branch=None,
        source_depth=None,
        silent=False,
        source_checksum=None,
    ):
        super().__init__(
            source,
            source_dir,
            source_tag,
            source_commit,
            source_branch,
            source_depth,
            source_checksum,
            "git",
        )
        if source_tag and source_branch:
            raise errors.SnapcraftSourceIncompatibleOptionsError(
                "git", ["source-tag", "source-branch"]
            )
        if source_tag and source_commit:
            raise errors.SnapcraftSourceIncompatibleOptionsError(
                "git", ["source-tag", "source-commit"]
            )
        if source_branch and source_commit:
            raise errors.SnapcraftSourceIncompatibleOptionsError(
                "git", ["source-branch", "source-commit"]
            )
        if source_checksum:
            raise errors.SnapcraftSourceInvalidOptionError("git", "source-checksum")
        self._call_kwargs = {}
        if silent:
            self._call_kwargs["stdout"] = subprocess.DEVNULL
            self._call_kwargs["stderr"] = subprocess.DEVNULL

    def _run_git_command(self, command: List[str]) -> None:
        try:
            subprocess.check_output(command)
        except subprocess.CalledProcessError as error:
            raise errors.GitCommandError(
                command=command,
                exit_code=error.returncode,
                output=error.output.decode(),
            )

    def _fetch_origin_commit(self):
        self._run(
            [
                self.command,
                "-C",
                self.source_dir,
                "fetch",
                "origin",
                self.source_commit,
            ],
            **self._call_kwargs
        )

    def _pull_existing(self):
        refspec = "HEAD"
        if self.source_branch:
            refspec = "refs/heads/" + self.source_branch
        elif self.source_tag:
            refspec = "refs/tags/" + self.source_tag
        elif self.source_commit:
            refspec = self.source_commit
            self._fetch_origin_commit()

        reset_spec = refspec if refspec != "HEAD" else "origin/master"

        self._run(
            [
                self.command,
                "-C",
                self.source_dir,
                "fetch",
                "--prune",
                "--recurse-submodules=yes",
            ],
            **self._call_kwargs
        )

        self._run(
            [self.command, "-C", self.source_dir, "reset", "--hard", reset_spec],
            **self._call_kwargs
        )

        # Merge any updates for the submodules (if any).
        self._run(
            [
                self.command,
                "-C",
                self.source_dir,
                "submodule",
                "update",
                "--recursive",
                "--force",
            ],
            **self._call_kwargs
        )

    def _clone_new(self):
        command = [self.command, "clone", "--recursive"]
        if self.source_tag or self.source_branch:
            command.extend(["--branch", self.source_tag or self.source_branch])
        if self.source_depth:
            command.extend(["--depth", str(self.source_depth)])
        self._run(command + [self.source, self.source_dir], **self._call_kwargs)

        if self.source_commit:
            self._fetch_origin_commit()

            self._run(
                [self.command, "-C", self.source_dir, "checkout", self.source_commit],
                **self._call_kwargs
            )

    def is_local(self):
        return os.path.exists(os.path.join(self.source_dir, ".git"))

    def pull(self):
        if self.is_local():
            self._pull_existing()
        else:
            self._clone_new()
        self.source_details = self._get_source_details()

    def push(self, url, refspec, force=False):
        command = [self.command, "-C", self.source_dir, "push", url, refspec]

        if force:
            command.append("--force")

        self._run_git_command(command)

    def init(self):
        command = [self.command, "-C", self.source_dir, "init"]
        self._run_git_command(command)

    def add(self, file):
        if file.startswith(self.source_dir):
            file = os.path.relpath(file, self.source_dir)

        command = [self.command, "-C", self.source_dir, "add", file]
        self._run_git_command(command)

    def commit(self, message, author="snapcraft <snapcraft@snapcraft.local>"):
        command = [
            self.command,
            "-C",
            self.source_dir,
            "commit",
            "--no-gpg-sign",
            "--message",
            message,
            "--author",
            author,
        ]
        self._run_git_command(command)

    def _get_source_details(self):
        tag = self.source_tag
        commit = self.source_commit
        branch = self.source_branch
        source = self.source
        checksum = self.source_checksum

        if not tag and not branch and not commit:
            commit = self._run_output(
                ["git", "-C", self.source_dir, "rev-parse", "HEAD"]
            )

        return {
            "source-commit": commit,
            "source-branch": branch,
            "source": source,
            "source-tag": tag,
            "source-checksum": checksum,
        }



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2019-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
import subprocess
from unittest import mock

import fixtures
import pytest
from testtools.matchers import Contains, Equals, HasLength

from snapcraft.plugins.v1._python import _pip, errors

from ._basesuite import PythonBaseTestCase


class PipRunBaseTestCase(PythonBaseTestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("snapcraft.internal.common.run_output")
        self.mock_run_output = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("snapcraft.internal.common.run")
        self.mock_run = patcher.start()
        self.addCleanup(patcher.stop)

        # Return json indicating that wheel and setuptools are installed
        self.mock_run_output.return_value = (
            '[{"name": "wheel", "version": "1.0"},'
            '{"name": "setuptools", "version": "1.0"}]'
        )

    def _assert_expected_enviroment(self, expected_python, headers_path):
        _pip.Pip(
            python_major_version="test",
            part_dir="part_dir",
            install_dir="install_dir",
            stage_dir="stage_dir",
        ).setup()

        class check_env:
            def __init__(self, test):
                self.test = test

            def __eq__(self, env):
                self.test.assertThat(env, Contains("PYTHONUSERBASE"))
                self.test.assertThat(env["PYTHONUSERBASE"], Equals("install_dir"))

                self.test.assertThat(env, Contains("PYTHONHOME"))
                if expected_python.startswith("install_dir"):
                    self.test.assertThat(
                        env["PYTHONHOME"], Equals(os.path.join("install_dir", "usr"))
                    )
                else:
                    self.test.assertThat(
                        env["PYTHONHOME"], Equals(os.path.join("stage_dir", "usr"))
                    )

                self.test.assertThat(env, Contains("PATH"))
                self.test.assertThat(
                    env["PATH"], Contains(os.path.join("install_dir", "usr", "bin"))
                )

                if headers_path:
                    self.test.assertThat(env, Contains("CPPFLAGS"))
                    self.test.assertThat(
                        env["CPPFLAGS"], Contains("-I{}".format(headers_path))
                    )

                return True

        self.mock_run_output.assert_has_calls(
            [
                mock.call(
                    [expected_python, "-m", "pip"],
                    stderr=subprocess.STDOUT,
                    env=check_env(self),
                ),
                mock.call(
                    [expected_python, "-m", "pip", "list", "--format=json"],
                    env=mock.ANY,
                ),
                mock.call(
                    [expected_python, "-m", "pip", "list", "--format=json"],
                    env=mock.ANY,
                ),
            ]
        )


class PipRunTestCase(PipRunBaseTestCase):
    def test_environment_part_python_without_headers(self):
        expected_python = self._create_python_binary("install_dir")
        self._assert_expected_enviroment(expected_python, None)

    def test_environment_part_python_with_staged_headers(self):
        expected_python = self._create_python_binary("install_dir")
        # First, create staged headers
        staged_headers = os.path.join("stage_dir", "usr", "include", "pythontest")
        os.makedirs(staged_headers)
        self._assert_expected_enviroment(expected_python, staged_headers)

    @mock.patch("glob.glob")
    def test_environment_part_python_with_host_headers(self, mock_glob):
        host_headers = os.path.join(os.sep, "usr", "include", "pythontest")

        # Fake out glob so it looks like the headers are installed on the host
        def _fake_glob(pattern):
            if pattern.startswith(os.sep):
                return [host_headers]
            return []

        mock_glob.side_effect = _fake_glob

        expected_python = self._create_python_binary("install_dir")
        self._assert_expected_enviroment(expected_python, host_headers)

    def test_environment_staged_python_without_headers(self):
        expected_python = self._create_python_binary("stage_dir")
        self._assert_expected_enviroment(expected_python, None)

    def test_environment_staged_python_with_staged_headers(self):
        # First, create staged headers
        staged_headers = os.path.join("stage_dir", "usr", "include", "pythontest")
        os.makedirs(staged_headers)

        # Also create staged python
        expected_python = self._create_python_binary("stage_dir")

        self._assert_expected_enviroment(expected_python, staged_headers)

    @mock.patch("glob.glob")
    def test_environment_staged_python_with_host_headers(self, mock_glob):
        host_headers = os.path.join(os.sep, "usr", "include", "pythontest")

        # Fake out glob so it looks like the headers are installed on the host
        def _fake_glob(pattern):
            if pattern.startswith(os.sep):
                return [host_headers]
            return []

        mock_glob.side_effect = _fake_glob

        expected_python = self._create_python_binary("stage_dir")
        self._assert_expected_enviroment(expected_python, host_headers)

    def test_with_extra_cppflags(self):
        """Verify that existing CPPFLAGS are preserved"""

        expected_python = self._create_python_binary("install_dir")

        self.useFixture(fixtures.EnvironmentVariable("CPPFLAGS", "-I/opt/include"))
        _pip.Pip(
            python_major_version="test",
            part_dir="part_dir",
            install_dir="install_dir",
            stage_dir="stage_dir",
        ).setup()

        class check_env:
            def __init__(self, test):
                self.test = test

            def __eq__(self, env):
                self.test.assertThat(env, Contains("CPPFLAGS"))
                self.test.assertThat(env["CPPFLAGS"], Contains("-I/opt/include"))

                return True

        self.mock_run_output.assert_has_calls(
            [
                mock.call(
                    [expected_python, "-m", "pip"],
                    env=check_env(self),
                    stderr=subprocess.STDOUT,
                )
            ]
        )


class SetupTestCase(PipRunBaseTestCase):
    def setUp(self):
        super().setUp()

        self.command = [self._create_python_binary("install_dir"), "-m", "pip"]

    def _assert_check_for_pip(self):
        self.mock_run_output.assert_has_calls(
            [
                mock.call(self.command, stderr=subprocess.STDOUT, env=mock.ANY),
                mock.call(self.command + ["list", "--format=json"], env=mock.ANY),
                mock.call(self.command + ["list", "--format=json"], env=mock.ANY),
            ]
        )

    def test_setup_with_pip_installed(self):
        """Test that no attempt is made to reinstall pip"""

        # Since _run doesn't raise an exception indicating pip isn't installed,
        # it must be installed.

        # Verify that no attempt is made to reinstall pip
        _pip.Pip(
            python_major_version="test",
            part_dir="part_dir",
            install_dir="install_dir",
            stage_dir="stage_dir",
        ).setup()

        self._assert_check_for_pip()
        self.mock_run.assert_not_called()

    def test_setup_without_pip_installed(self):
        """Test that the system pip is used to install our own pip"""

        # Raise an exception indicating that pip isn't installed
        def fake_run(command, **kwargs):
            if command == self.command:
                raise subprocess.CalledProcessError(1, "foo", b"no module named pip")

            return "[]"

        self.mock_run_output.side_effect = fake_run

        # Verify that pip is then installed
        _pip.Pip(
            python_major_version="test",
            part_dir="part_dir",
            install_dir="install_dir",
            stage_dir="stage_dir",
        ).setup()

        part_pythonhome = os.path.join("install_dir", "usr")
        host_pythonhome = os.path.join(os.path.sep, "usr")

        # What we're asserting here:
        # 1. That we test for the installed pip
        # 2. That we then download pip (and associated tools) using host pip
        # 3. That we then install pip (and associated tools) using host pip

        # Check that we correctly test for the installed pip
        self.mock_run_output.assert_has_calls(
            [
                mock.call(
                    self.command,
                    stderr=subprocess.STDOUT,
                    env=_CheckPythonhomeEnv(self, part_pythonhome),
                ),
                mock.call(self.command + ["list", "--format=json"], env=mock.ANY),
                mock.call(self.command + ["list", "--format=json"], env=mock.ANY),
            ]
        )

        # Now test that we download and install pip using the host's pip, and
        # then install the associated tools using the pip we pulled.
        self.assertThat(self.mock_run.mock_calls, HasLength(6))
        self.mock_run.assert_has_calls(
            [
                mock.call(
                    _CheckCommand(self, "download", ["pip"], []),
                    env=_CheckPythonhomeEnv(self, host_pythonhome),
                    cwd=None,
                ),
                mock.call(
                    _CheckCommand(self, "install", ["pip"], ["--ignore-installed"]),
                    env=_CheckPythonhomeEnv(self, host_pythonhome),
                    cwd=None,
                ),
                mock.call(
                    _CheckCommand(self, "download", ["wheel"], []),
                    env=_CheckPythonhomeEnv(self, part_pythonhome),
                    cwd=None,
                ),
                mock.call(
                    _CheckCommand(self, "install", ["wheel"], ["--ignore-installed"]),
                    env=_CheckPythonhomeEnv(self, part_pythonhome),
                    cwd=None,
                ),
                mock.call(
                    _CheckCommand(self, "download", ["setuptools"], []),
                    env=_CheckPythonhomeEnv(self, part_pythonhome),
                    cwd=None,
                ),
                mock.call(
                    _CheckCommand(
                        self, "install", ["setuptools"], ["--ignore-installed"]
                    ),
                    env=_CheckPythonhomeEnv(self, part_pythonhome),
                    cwd=None,
                ),
            ]
        )

    def test_setup_unexpected_error(self):
        """Test that pip initialization doesn't eat legit errors"""

        # Raises an exception indicating something bad happened
        self.mock_run_output.side_effect = subprocess.CalledProcessError(
            1, "foo", b"no good, very bad"
        )

        pip = _pip.Pip(
            python_major_version="test",
            part_dir="part_dir",
            install_dir="install_dir",
            stage_dir="stage_dir",
        )

        # Verify that pip lets that exception through
        self.assertRaises(subprocess.CalledProcessError, pip.setup)


class PipBaseTestCase(PythonBaseTestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch.object(_pip.Pip, "_run")
        self.mock_run = patcher.start()
        self.addCleanup(patcher.stop)

        self._create_python_binary("install_dir")


class PipTestCase(PipBaseTestCase):
    def test_clean_packages(self):
        pip = _pip.Pip(
            python_major_version="test",
            part_dir="part_dir",
            install_dir="install_dir",
            stage_dir="stage_dir",
        )

        packages_dir = os.path.join("part_dir", "python-packages")
        self.assertTrue(os.path.exists(packages_dir))

        # Now verify that asking pip to clean removes its packages
        pip.clean_packages()
        self.assertFalse(os.path.exists(packages_dir))


class PipCommandBaseTestCase(PipBaseTestCase):
    def setUp(self):
        super().setUp()

        self.pip = _pip.Pip(
            python_major_version="test",
            part_dir="part_dir",
            install_dir="install_dir",
            stage_dir="stage_dir",
        )

        # We don't care about anything init did to the mock here: reset it
        self.mock_run.reset_mock()


@pytest.fixture
def mock_pip_run():
    patcher = mock.patch.object(_pip.Pip, "_run")
    yield patcher.start()
    patcher.stop()


@pytest.fixture
def pip_instance(tmp_work_path):
    python_command_path = tmp_work_path / "install_dir/usr/bin/pythontest"
    python_command_path.parent.mkdir(parents=True)
    python_command_path.touch()

    return _pip.Pip(
        python_major_version="test",
        part_dir="part_dir",
        install_dir="install_dir",
        stage_dir="stage_dir",
    )


class TestPipDownload:

    scenarios = [
        (
            "packages",
            {
                "packages": ["foo", "bar"],
                "kwargs": {},
                "expected_args": ["foo", "bar"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "setup_py_dir",
            {
                "packages": [],
                "kwargs": {"setup_py_dir": "test_setup_py_dir"},
                "expected_args": ["."],
                "expected_kwargs": {"cwd": "test_setup_py_dir"},
            },
        ),
        (
            "single constraint",
            {
                "packages": ["foo"],
                "kwargs": {"constraints": ["constraint"]},
                "expected_args": ["--constraint", "constraint", "foo"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "multiple constraints",
            {
                "packages": ["foo"],
                "kwargs": {"constraints": ["constraint1", "constraint2"]},
                "expected_args": [
                    "--constraint",
                    "constraint1",
                    "--constraint",
                    "constraint2",
                    "foo",
                ],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "single requirement",
            {
                "packages": [],
                "kwargs": {"requirements": ["requirement"]},
                "expected_args": ["--requirement", "requirement"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "multiple requirements",
            {
                "packages": [],
                "kwargs": {"requirements": ["requirement1", "requirement2"]},
                "expected_args": [
                    "--requirement",
                    "requirement1",
                    "--requirement",
                    "requirement2",
                ],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "process dependency links",
            {
                "packages": ["foo"],
                "kwargs": {"process_dependency_links": True},
                "expected_args": ["--process-dependency-links", "foo"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "packages and setup_py_dir",
            {
                "packages": ["foo", "bar"],
                "kwargs": {"setup_py_dir": "test_setup_py_dir"},
                "expected_args": ["foo", "bar", "."],
                "expected_kwargs": {"cwd": "test_setup_py_dir"},
            },
        ),
    ]

    def test_with_packages_or_setup_py_or_requirements(
        self,
        pip_instance,
        mock_pip_run,
        packages,
        kwargs,
        expected_args,
        expected_kwargs,
    ):
        pip_instance.download(packages, **kwargs)

        args = [
            "download",
            "--disable-pip-version-check",
            "--dest",
            mock.ANY,
        ] + expected_args
        mock_pip_run.assert_called_once_with(args, **expected_kwargs)


def test_download_without_packages_or_setup_py_or_requirements_should_noop(
    mock_pip_run, pip_instance
):
    pip_instance.download([])

    mock_pip_run.assert_not_called()


class TestPipInstall:

    scenarios = [
        (
            "packages",
            {
                "packages": ["foo", "bar"],
                "kwargs": {},
                "expected_args": ["--no-index", "foo", "bar"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "setup_py_dir",
            {
                "packages": [],
                "kwargs": {"setup_py_dir": "test_setup_py_dir"},
                "expected_args": ["--no-index", "."],
                "expected_kwargs": {"cwd": "test_setup_py_dir"},
            },
        ),
        (
            "single constraint",
            {
                "packages": ["foo"],
                "kwargs": {"constraints": ["constraint"]},
                "expected_args": ["--constraint", "constraint", "--no-index", "foo"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "multiple constraints",
            {
                "packages": ["foo"],
                "kwargs": {"constraints": ["constraint1", "constraint2"]},
                "expected_args": [
                    "--constraint",
                    "constraint1",
                    "--constraint",
                    "constraint2",
                    "--no-index",
                    "foo",
                ],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "single requirement",
            {
                "packages": [],
                "kwargs": {"requirements": ["requirement"]},
                "expected_args": ["--no-index", "--requirement", "requirement"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "multiple requirements",
            {
                "packages": [],
                "kwargs": {"requirements": ["requirement1", "requirement2"]},
                "expected_args": [
                    "--no-index",
                    "--requirement",
                    "requirement1",
                    "--requirement",
                    "requirement2",
                ],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "process dependency links",
            {
                "packages": ["foo"],
                "kwargs": {"process_dependency_links": True},
                "expected_args": ["--process-dependency-links", "--no-index", "foo"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "use PyPI",
            {
                "packages": [],
                "kwargs": {"setup_py_dir": "test_setup_py_dir", "no_index": False},
                "expected_args": ["."],
                "expected_kwargs": {"cwd": "test_setup_py_dir"},
            },
        ),
        (
            "upgrade",
            {
                "packages": ["foo"],
                "kwargs": {"upgrade": True},
                "expected_args": ["--upgrade", "--no-index", "foo"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "no install_deps",
            {
                "packages": ["foo"],
                "kwargs": {"install_deps": False},
                "expected_args": ["--no-deps", "--no-index", "foo"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "ignore_installed",
            {
                "packages": ["foo"],
                "kwargs": {"ignore_installed": True},
                "expected_args": ["--ignore-installed", "--no-index", "foo"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "packages and setup_py_dir",
            {
                "packages": ["foo", "bar"],
                "kwargs": {"setup_py_dir": "test_setup_py_dir"},
                "expected_args": ["--no-index", "foo", "bar", "."],
                "expected_kwargs": {"cwd": "test_setup_py_dir"},
            },
        ),
    ]

    def test_with_packages_or_setup_py_or_requirements(
        self,
        mock_pip_run,
        pip_instance,
        packages,
        kwargs,
        expected_args,
        expected_kwargs,
    ):
        pip_instance.install(packages, **kwargs)

        args = [
            "install",
            "--user",
            "--no-compile",
            "--find-links",
            mock.ANY,
        ] + expected_args
        mock_pip_run.assert_called_once_with(args, **expected_kwargs)


def test_install_without_packages_or_setup_py_or_requirements_should_noop(
    mock_pip_run, pip_instance
):
    pip_instance.install([])

    mock_pip_run.assert_not_called()


class TestPipInstallFixupShebang:

    scenarios = [
        (
            "bad shebang",
            {
                "file_path": "example.py",
                "contents": "#!/foo/bar/baz/python",
                "expected": "#!/usr/bin/env python",
            },
        ),
        (
            "bad shebang and bad permissions",
            {
                "file_path": os.path.join("bin", "another_example.py"),
                "contents": "#!/foo/baz/python3",
                "expected": "#!/usr/bin/env python3",
            },
        ),
        (
            "no shebang",
            {
                "file_path": "foo",
                "contents": "foo\n#!/usr/bin/python3",
                "expected": "foo\n#!/usr/bin/python3",
            },
        ),
        (
            "no shebang and bad permissions",
            {"file_path": "bar", "contents": "bar", "expected": "bar"},
        ),
        (
            "good shebang",
            {
                "file_path": "baz",
                "contents": "#!/usr/bin/env python",
                "expected": "#!/usr/bin/env python",
            },
        ),
        (
            "good shebang and bad permissions",
            {
                "file_path": os.path.join("qux", "another_example.py"),
                "contents": "#!/usr/bin/env python",
                "expected": "#!/usr/bin/env python",
            },
        ),
    ]

    def test_install_fixes_shebangs(
        self, mock_pip_run, pip_instance, file_path, contents, expected
    ):
        test_file_path = pathlib.Path(pip_instance._install_dir) / file_path
        test_file_path.parent.mkdir(parents=True, exist_ok=True)

        with test_file_path.open("w") as test_file:
            print(contents, end="", file=test_file)

        pip_instance.install(["foo"])

        with test_file_path.open() as test_file:
            assert test_file.read() == expected


class TestPipInstallFixupPermissions:

    scenarios = [
        ("755", {"file_path": "example.py", "mode": 0o755, "expected_mode": "755"}),
        (
            "500",
            {
                "file_path": os.path.join("bin", "another_example.py"),
                "mode": 0o500,
                "expected_mode": "755",
            },
        ),
        ("777", {"file_path": "foo", "mode": 0o777, "expected_mode": "777"}),
        ("700", {"file_path": "bar", "mode": 0o700, "expected_mode": "755"}),
        (
            "000",
            {
                "file_path": os.path.join("qux", "another_example.py"),
                "mode": 0o000,
                "expected_mode": "200",
            },
        ),
    ]

    def test_install_fixes_permissions(
        self, mock_pip_run, pip_instance, file_path, mode, expected_mode
    ):
        test_file_path = pathlib.Path(pip_instance._install_dir) / file_path
        test_file_path.parent.mkdir(parents=True, exist_ok=True)
        test_file_path.touch()
        test_file_path.chmod(mode)

        pip_instance.install(["foo"])

        assert oct(test_file_path.stat().st_mode)[-3:] == expected_mode


@mock.patch.object(os, "chmod")
def test_symlink(mock_chmod, tmp_work_path):
    test_file_path = tmp_work_path / "file"
    test_file_path.touch()
    pathlib.Path("link").symlink_to(test_file_path)

    _pip._replicate_owner_mode("link")

    assert mock_chmod.called is False


class TestPipWheel:

    scenarios = [
        (
            "packages",
            {
                "packages": ["foo", "bar"],
                "kwargs": {},
                "expected_args": ["foo", "bar"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "setup_py_dir",
            {
                "packages": [],
                "kwargs": {"setup_py_dir": "test_setup_py_dir"},
                "expected_args": ["."],
                "expected_kwargs": {"cwd": "test_setup_py_dir"},
            },
        ),
        (
            "single constraint",
            {
                "packages": ["foo"],
                "kwargs": {"constraints": ["constraint"]},
                "expected_args": ["--constraint", "constraint", "foo"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "multiple constraints",
            {
                "packages": ["foo"],
                "kwargs": {"constraints": ["constraint1", "constraint2"]},
                "expected_args": [
                    "--constraint",
                    "constraint1",
                    "--constraint",
                    "constraint2",
                    "foo",
                ],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "single requirement",
            {
                "packages": [],
                "kwargs": {"requirements": ["requirement"]},
                "expected_args": ["--requirement", "requirement"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "multiple requirements",
            {
                "packages": [],
                "kwargs": {"requirements": ["requirement1", "requirement2"]},
                "expected_args": [
                    "--requirement",
                    "requirement1",
                    "--requirement",
                    "requirement2",
                ],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "process dependency links",
            {
                "packages": ["foo"],
                "kwargs": {"process_dependency_links": True},
                "expected_args": ["--process-dependency-links", "foo"],
                "expected_kwargs": {"cwd": None},
            },
        ),
        (
            "packages and setup_py_dir",
            {
                "packages": ["foo", "bar"],
                "kwargs": {"setup_py_dir": "test_setup_py_dir"},
                "expected_args": ["foo", "bar", "."],
                "expected_kwargs": {"cwd": "test_setup_py_dir"},
            },
        ),
    ]

    def test_with_packages_or_setup_py_or_requirements(
        mock_common_run_output,
        mock_pip_run,
        pip_instance,
        packages,
        kwargs,
        expected_args,
        expected_kwargs,
    ):
        mock_common_run_output.return_value = [
            '[{"name": "wheel", "version": "1.0"},'
            '{"name": "setuptools", "version": "1.0"}]'
        ]

        pip_instance.wheel(packages, **kwargs)

        args = [
            "wheel",
            "--no-index",
            "--find-links",
            mock.ANY,
            "--wheel-dir",
            mock.ANY,
        ] + expected_args
        mock_pip_run.assert_called_once_with(args, **expected_kwargs)


def test_wheel_without_packages_or_setup_py_or_requirements_should_noop(
    mock_pip_run, pip_instance
):
    pip_instance.wheel([])

    mock_pip_run.assert_not_called()


class PipListTestCase(PipCommandBaseTestCase):
    def test_none(self):
        self.mock_run.return_value = "{}"
        self.assertFalse(self.pip.list())
        self.mock_run.assert_called_once_with(
            ["list", "--format=json"], runner=mock.ANY
        )

    def test_package(self):
        self.mock_run.return_value = '[{"name": "foo", "version": "1.0"}]'
        self.assertThat(self.pip.list(), Equals({"foo": "1.0"}))
        self.mock_run.assert_called_once_with(
            ["list", "--format=json"], runner=mock.ANY
        )

    def test_user(self):
        self.mock_run.return_value = '[{"name": "foo", "version": "1.0"}]'
        self.assertThat(self.pip.list(user=True), Equals({"foo": "1.0"}))
        self.mock_run.assert_called_once_with(
            ["list", "--user", "--format=json"], runner=mock.ANY
        )

    def test_missing_name(self):
        self.mock_run.return_value = '[{"version": "1.0"}]'
        raised = self.assertRaises(errors.PipListMissingFieldError, self.pip.list)
        self.assertThat(str(raised), Contains("Pip packages json missing 'name' field"))

    def test_missing_version(self):
        self.mock_run.return_value = '[{"name": "foo"}]'
        raised = self.assertRaises(errors.PipListMissingFieldError, self.pip.list)
        self.assertThat(
            str(raised), Contains("Pip packages json missing 'version' field")
        )

    def test_invalid_json(self):
        self.mock_run.return_value = "[{]"
        raised = self.assertRaises(errors.PipListInvalidJsonError, self.pip.list)
        self.assertThat(str(raised), Contains("Pip packages output isn't valid json"))

    def test_no_json_format(self):
        self.mock_run.side_effect = [
            subprocess.CalledProcessError(2, "pip", "no such option"),
            "foo (1.0)",
        ]
        self.assertThat(self.pip.list(), Equals({"foo": "1.0"}))
        self.mock_run.assert_has_calls(
            [
                mock.call(["list", "--format=json"], runner=mock.ANY),
                mock.call(["list"], runner=mock.ANY),
            ]
        )

    def test_no_json_format_invalid_result(self):
        self.mock_run.side_effect = [
            subprocess.CalledProcessError(2, "pip", "no such option"),
            "foo 1.0",
        ]
        raised = self.assertRaises(
            errors.PipListInvalidLegacyFormatError, self.pip.list
        )
        self.assertThat(raised.output, Equals("foo 1.0"))


class _CheckPythonhomeEnv:
    def __init__(self, test, expected_pythonhome):
        self.test = test
        self.expected_pythonhome = expected_pythonhome

    def __eq__(self, env):
        # Verify that we're using the installed pip
        self.test.assertThat(env, Contains("PYTHONHOME"))
        self.test.assertThat(env["PYTHONHOME"], Equals(self.expected_pythonhome))

        return True


class _CheckCommand:
    def __init__(self, test, command, packages, flags):
        self.test = test
        self.command = command
        self.packages = packages
        self.flags = flags

    def __eq__(self, command):
        # Not worrying about the command arguments here, those are
        # tested elsewhere. Just want to test that the right command
        # is called with the right packages.
        self.test.assertTrue(command)
        self.test.assertThat(command[len(self.test.command)], Equals(self.command))

        for package in self.packages:
            self.test.assertThat(command, Contains(package))

        for flag in self.flags:
            self.test.assertThat(command, Contains(flag))

        return True



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testtools.matchers import Contains, Equals

from snapcraft import storeapi
from snapcraft.storeapi.v2.channel_map import (
    MappedChannel,
    Progressive,
    Revision,
    SnapChannel,
)

from . import FakeStoreCommandsBaseTestCase


class StatusCommandTestCase(FakeStoreCommandsBaseTestCase):
    def test_status_without_snap_raises_exception(self):
        result = self.run_command(["status"])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("Usage:"))

    def test_status_without_login_must_ask(self):
        self.fake_store_get_snap_channel_map.mock.side_effect = [
            storeapi.http_clients.errors.InvalidCredentialsError("error"),
            self.channel_map,
        ]

        result = self.run_command(
            ["status", "snap-test"], input="user@example.com\nsecret\n"
        )
        self.assertThat(
            result.output, Contains("You are required to login before continuing.")
        )

    def test_status(self):
        result = self.run_command(["status", "snap-test"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Track    Arch    Channel    Version    Revision
            2.1      amd64   stable     -          -
                             candidate  -          -
                             beta       10         19
                             edge       ↑          ↑
            2.0      amd64   stable     -          -
                             candidate  -          -
                             beta       10         18
                             edge       ↑          ↑
            """
                )
            ),
        )

    def test_status_following(self):
        self.channel_map.channel_map = [
            MappedChannel(
                channel="2.1/stable",
                architecture="amd64",
                expiration_date="2020-02-03T20:58:37Z",
                revision=20,
                progressive=Progressive(
                    paused=None, percentage=None, current_percentage=None
                ),
            )
        ]
        self.channel_map.revisions.append(
            Revision(architectures=["amd64"], revision=20, version="10")
        )

        result = self.run_command(["status", "snap-test"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Track    Arch    Channel    Version    Revision
            2.1      amd64   stable     10         20
                             candidate  ↑          ↑
                             beta       ↑          ↑
                             edge       ↑          ↑
            """
                )
            ),
        )

    def test_status_no_releases(self):
        self.channel_map.channel_map = []

        result = self.run_command(["status", "snap-test"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output.strip(), Equals("This snap has no released revisions.")
        )

    def test_progressive_status(self):
        self.channel_map.channel_map.append(
            MappedChannel(
                channel="2.1/beta",
                architecture="amd64",
                expiration_date="2020-02-03T20:58:37Z",
                revision=20,
                progressive=Progressive(
                    paused=None, percentage=10.0, current_percentage=7.2
                ),
            )
        )
        self.channel_map.revisions.append(
            Revision(architectures=["amd64"], revision=20, version="11")
        )

        result = self.run_command(
            ["status", "snap-test", "--experimental-progressive-releases"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            *EXPERIMENTAL* progressive releases in use.
            Track    Arch    Channel    Version    Revision    Progress
            2.1      amd64   stable     -          -           -
                             candidate  -          -           -
                             beta       10         19          93→90%
                                        11         20          7→10%
                             edge       ↑          ↑           -
            2.0      amd64   stable     -          -           -
                             candidate  -          -           -
                             beta       10         18          -
                             edge       ↑          ↑           -
            """
                )
            ),
        )

    def test_status_by_arch(self):
        self.channel_map.channel_map.append(
            MappedChannel(
                channel="2.1/beta",
                architecture="s390x",
                expiration_date=None,
                revision=99,
                progressive=Progressive(
                    paused=None, percentage=None, current_percentage=None
                ),
            )
        )
        self.channel_map.revisions.append(
            Revision(architectures=["s390x"], revision=99, version="10")
        )

        result = self.run_command(["status", "snap-test", "--arch=s390x"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Track    Arch    Channel    Version    Revision
            2.1      s390x   stable     -          -
                             candidate  -          -
                             beta       10         99
                             edge       ↑          ↑
            """
                )
            ),
        )

    def test_status_by_multiple_arch(self):
        self.channel_map.channel_map.append(
            MappedChannel(
                channel="2.1/beta",
                architecture="s390x",
                expiration_date=None,
                revision=98,
                progressive=Progressive(
                    paused=None, percentage=None, current_percentage=None
                ),
            )
        )
        self.channel_map.channel_map.append(
            MappedChannel(
                channel="2.1/beta",
                architecture="arm64",
                expiration_date=None,
                revision=99,
                progressive=Progressive(
                    paused=None, percentage=None, current_percentage=None
                ),
            )
        )
        self.channel_map.revisions.append(
            Revision(architectures=["s390x"], revision=98, version="10")
        )
        self.channel_map.revisions.append(
            Revision(architectures=["arm64"], revision=99, version="10")
        )

        result = self.run_command(
            ["status", "snap-test", "--arch=s390x", "--arch=arm64"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Track    Arch    Channel    Version    Revision
            2.1      arm64   stable     -          -
                             candidate  -          -
                             beta       10         99
                             edge       ↑          ↑
                     s390x   stable     -          -
                             candidate  -          -
                             beta       10         98
                             edge       ↑          ↑
            """
                )
            ),
        )

    def test_status_by_track(self):
        result = self.run_command(["status", "snap-test", "--track=2.0"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Track    Arch    Channel    Version    Revision
            2.0      amd64   stable     -          -
                             candidate  -          -
                             beta       10         18
                             edge       ↑          ↑
            """
                )
            ),
        )

    def test_status_by_multiple_track(self):
        result = self.run_command(["status", "snap-test", "--track=2.0", "--track=2.1"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Track    Arch    Channel    Version    Revision
            2.1      amd64   stable     -          -
                             candidate  -          -
                             beta       10         19
                             edge       ↑          ↑
            2.0      amd64   stable     -          -
                             candidate  -          -
                             beta       10         18
                             edge       ↑          ↑
            """
                )
            ),
        )

    def test_status_by_track_and_arch(self):
        self.channel_map.channel_map.append(
            MappedChannel(
                channel="2.1/beta",
                architecture="s390x",
                expiration_date=None,
                revision=99,
                progressive=Progressive(
                    paused=None, percentage=None, current_percentage=None
                ),
            )
        )
        self.channel_map.channel_map.append(
            MappedChannel(
                channel="2.0/beta",
                architecture="s390x",
                expiration_date=None,
                revision=98,
                progressive=Progressive(
                    paused=None, percentage=None, current_percentage=None
                ),
            )
        )
        self.channel_map.revisions.append(
            Revision(architectures=["s390x"], revision=99, version="10")
        )
        self.channel_map.revisions.append(
            Revision(architectures=["s390x"], revision=98, version="10")
        )

        result = self.run_command(
            ["status", "snap-test", "--arch=s390x", "--track=2.0"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Track    Arch    Channel    Version    Revision
            2.0      s390x   stable     -          -
                             candidate  -          -
                             beta       10         98
                             edge       ↑          ↑
            """
                )
            ),
        )

    def test_status_including_branch(self):
        self.channel_map.channel_map.append(
            MappedChannel(
                channel="2.1/stable/hotfix1",
                architecture="amd64",
                expiration_date="2020-02-03T20:58:37Z",
                revision=20,
                progressive=Progressive(
                    paused=None, percentage=None, current_percentage=None
                ),
            )
        )
        self.channel_map.revisions.append(
            Revision(architectures=["amd64"], revision=20, version="10hotfix")
        )
        self.channel_map.snap.channels.append(
            SnapChannel(
                name="2.1/stable/hotfix1",
                track="2.1",
                risk="stable",
                branch="hotfix1",
                fallback="2.1/stable",
            )
        )

        result = self.run_command(["status", "snap-test"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            Track    Arch    Channel         Version    Revision    Expires at
            2.1      amd64   stable          -          -
                             stable/hotfix1  10hotfix   20          2020-02-03T20:58:37Z
                             candidate       -          -
                             beta            10         19
                             edge            ↑          ↑
            2.0      amd64   stable          -          -
                             candidate       -          -
                             beta            10         18
                             edge            ↑          ↑
            """
                )
            ),
        )

    def test_progressive_status_including_branch(self):
        self.channel_map.channel_map.append(
            MappedChannel(
                channel="2.1/stable/hotfix1",
                architecture="amd64",
                expiration_date="2020-02-03T20:58:37Z",
                revision=20,
                progressive=Progressive(
                    paused=None, percentage=20.0, current_percentage=12.3
                ),
            )
        )
        self.channel_map.revisions.append(
            Revision(architectures=["amd64"], revision=20, version="10hotfix")
        )
        self.channel_map.snap.channels.append(
            SnapChannel(
                name="2.1/stable/hotfix1",
                track="2.1",
                risk="stable",
                branch="hotfix1",
                fallback="2.1/stable",
            )
        )

        result = self.run_command(
            ["status", "snap-test", "--experimental-progressive-releases"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            *EXPERIMENTAL* progressive releases in use.
            Track    Arch    Channel         Version    Revision    Progress    Expires at
            2.1      amd64   stable          -          -           -
                             stable/hotfix1  10hotfix   20          12→20%      2020-02-03T20:58:37Z
                             candidate       -          -           -
                             beta            10         19          -
                             edge            ↑          ↑           -
            2.0      amd64   stable          -          -           -
                             candidate       -          -           -
                             beta            10         18          -
                             edge            ↑          ↑           -
            """
                )
            ),
        )

    def test_progressive_status_with_null_current_percentage(self):
        self.channel_map.channel_map[0].progressive.percentage = 10.0
        self.channel_map.channel_map[0].progressive.current_percentage = None

        result = self.run_command(
            ["status", "snap-test", "--experimental-progressive-releases"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Equals(
                dedent(
                    """\
            *EXPERIMENTAL* progressive releases in use.
            Track    Arch    Channel    Version    Revision    Progress
            2.1      amd64   stable     -          -           -
                             candidate  -          -           -
                             beta       -          -           -
                                        10         19          ?→10%
                             edge       ↑          ↑           -
            2.0      amd64   stable     -          -           -
                             candidate  -          -           -
                             beta       10         18          -
                             edge       ↑          ↑           -
            """
                )
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import pathlib
from textwrap import dedent

import fixtures
import pytest
from testscenarios.scenarios import multiply_scenarios
from testtools.matchers import Contains, Equals

import snapcraft.yaml_utils.errors
from snapcraft import project
from snapcraft.internal.errors import PluginError
from snapcraft.internal.project_loader import errors, load_config
from tests import fixture_setup

from . import ProjectLoaderBaseTest


def get_project_config(snapcraft_yaml_content, target_deb_arch=None):
    snapcraft_yaml_path = pathlib.Path("Snapcraft.yaml")
    with snapcraft_yaml_path.open("w") as snapcraft_yaml_file:
        print(snapcraft_yaml_content, file=snapcraft_yaml_file)

    snapcraft_project = project.Project(
        snapcraft_yaml_file_path=snapcraft_yaml_path.as_posix(),
        target_deb_arch=target_deb_arch,
    )
    return load_config(snapcraft_project)


class TestValidArchitectures:

    yaml_scenarios = [
        (
            "none",
            {
                "expected_amd64": ["amd64"],
                "expected_i386": ["i386"],
                "expected_armhf": ["armhf"],
                "yaml": None,
            },
        ),
        (
            "single string list",
            {
                "expected_amd64": ["amd64"],
                "expected_i386": ["i386"],
                "expected_armhf": ["armhf"],
                "yaml": "[amd64]",
            },
        ),
        (
            "multiple string list",
            {
                "expected_amd64": ["amd64", "i386"],
                "expected_i386": ["amd64", "i386"],
                "expected_armhf": ["armhf"],
                "yaml": "[amd64, i386]",
            },
        ),
        (
            "single object list",
            {
                "expected_amd64": ["amd64"],
                "expected_i386": ["i386"],
                "expected_armhf": ["armhf"],
                "yaml": dedent(
                    """
                - build-on: [amd64]
                  run-on: [amd64]
            """
                ),
            },
        ),
        (
            "multiple object list",
            {
                "expected_amd64": ["amd64"],
                "expected_i386": ["i386"],
                "expected_armhf": ["armhf", "arm64"],
                "yaml": dedent(
                    """
                - build-on: [amd64]
                  run-on: [amd64]
                - build-on: [i386]
                  run-on: [i386]
                - build-on: [armhf]
                  run-on: [armhf, arm64]
            """
                ),
            },
        ),
        (
            "omit run-on",
            {
                "expected_amd64": ["amd64"],
                "expected_i386": ["i386"],
                "expected_armhf": ["armhf"],
                "yaml": dedent(
                    """
                - build-on: [amd64]
            """
                ),
            },
        ),
        (
            "single build-on string, no list",
            {
                "expected_amd64": ["amd64"],
                "expected_i386": ["i386"],
                "expected_armhf": ["armhf"],
                "yaml": dedent(
                    """
                - build-on: amd64
            """
                ),
            },
        ),
        (
            "build- and run-on string, no lists",
            {
                "expected_amd64": ["amd64"],
                "expected_i386": ["amd64"],
                "expected_armhf": ["armhf"],
                "yaml": dedent(
                    """
                - build-on: i386
                  run-on: amd64
            """
                ),
            },
        ),
        (
            "build on all",
            {
                "expected_amd64": ["amd64"],
                "expected_i386": ["amd64"],
                "expected_armhf": ["amd64"],
                "yaml": dedent(
                    """
                - build-on: [all]
                  run-on: [amd64]
            """
                ),
            },
        ),
        (
            "run on all",
            {
                "expected_amd64": ["all"],
                "expected_i386": ["i386"],
                "expected_armhf": ["armhf"],
                "yaml": dedent(
                    """
                - build-on: [amd64]
                  run-on: [all]
            """
                ),
            },
        ),
    ]

    arch_scenarios = [
        ("amd64", {"target_arch": "amd64"}),
        ("i386", {"target_arch": "i386"}),
        ("armhf", {"target_arch": "armhf"}),
    ]

    scenarios = multiply_scenarios(yaml_scenarios, arch_scenarios)

    def test(
        self,
        tmp_work_path,
        yaml,
        target_arch,
        expected_amd64,
        expected_i386,
        expected_armhf,
    ):
        snippet = "architectures: {}".format(yaml) if yaml else ""
        snapcraft_yaml = dedent(
            """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            {}
            parts:
              my-part:
                plugin: nil
        """
        ).format(snippet)

        c = get_project_config(snapcraft_yaml, target_deb_arch=target_arch)

        expected_targets = {
            "amd64": expected_amd64,
            "i386": expected_i386,
            "armhf": expected_armhf,
        }
        assert c.data["architectures"] == expected_targets[target_arch]


class AliasesTest(ProjectLoaderBaseTest):
    def make_snapcraft_project(self, apps):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part("part1", dict(plugin="nil"))
        for app_name, app in apps:
            snapcraft_yaml.update_app(app_name, app)
        self.useFixture(snapcraft_yaml)

        p = project.Project(
            snapcraft_yaml_file_path=snapcraft_yaml.snapcraft_yaml_file_path
        )
        return load_config(p)

    def test_aliases(self):
        fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(fake_logger)

        apps = [("test", dict(command="test", aliases=["test-it", "testing"]))]
        c = self.make_snapcraft_project(apps)

        self.maxDiff = None

        self.assertTrue(
            "aliases" in c.data["apps"]["test"],
            'Expected "aliases" property to be in snapcraft.yaml',
        )
        self.assertThat(
            c.data["apps"]["test"]["aliases"], Equals(["test-it", "testing"])
        )

        # Verify that aliases are properly deprecated
        self.assertThat(
            fake_logger.output,
            Contains(
                "Aliases are now handled by the store, and shouldn't be declared "
                "in the snap."
            ),
        )
        self.assertThat(
            fake_logger.output,
            Contains("See http://snapcraft.io/docs/deprecation-notices/dn5"),
        )

    def test_duplicate_aliases(self):
        apps = [
            ("test1", dict(command="test", aliases=["testing"])),
            ("test2", dict(command="test", aliases=["testing"])),
        ]
        raised = self.assertRaises(
            errors.DuplicateAliasError, self.make_snapcraft_project, apps
        )

        self.assertThat(
            str(raised),
            Equals(
                "Multiple parts have the same alias defined: {!r}".format("testing")
            ),
        )

    def test_invalid_alias(self):
        apps = [("test", dict(command="test", aliases=[".test"]))]
        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            self.make_snapcraft_project,
            apps,
        )
        expected = (
            "The {path!r} property does not match the required schema: "
            "{alias!r} does not match ".format(
                path="apps/test/aliases[0]", alias=".test"
            )
        )
        self.assertThat(str(raised), Contains(expected))


@pytest.mark.parametrize("entry", ["slots", "plugs"])
def test_loading_additional_properties_fails(tmp_work_path, entry):
    snapcraft_yaml = dedent(
        f"""\
            name: my-package-1
            base: core18
            version: 1.0-snapcraft1~ppa1
            summary: my summary less that 79 chars
            description: description which can be pretty long
            parts:
                part1:
                    plugin: nil
                    {entry}: [{entry}1]
        """
    )

    with pytest.raises(PluginError):
        get_project_config(snapcraft_yaml)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict

_BASE_IMAGE = dict(core="16.04", core16="16.04", core18="18.04", core20="20.04")


def get_image_source(*, base: str) -> Dict[str, str]:
    """Return a valid source to build for base.

    :param str base: the base to target.
    :returns: a dictionary with a valid source for PyLXD.
    :raises KeyError: when a non supported base is used.
    """
    # TODO catch KeyError and raise an appropriate exception.
    return dict(
        type="image",
        mode="pull",
        server="https://cloud-images.ubuntu.com/buildd/releases",
        protocol="simplestreams",
        alias=_BASE_IMAGE[base],
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Any, Dict, List, Set


class PluginV2(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def get_schema(cls) -> Dict[str, Any]:
        """Return a jsonschema compatible dictionary for the plugin properties."""

    def __init__(self, *, part_name: str, options) -> None:
        """
        :param str part_name: part names
        :param options: an object representing part defined properties.
        """
        self.name = part_name
        self.options = options

    @abc.abstractmethod
    def get_build_snaps(self) -> Set[str]:
        """
        Return a set of required packages to install in the build environment.
        """

    @abc.abstractmethod
    def get_build_packages(self) -> Set[str]:
        """
        Return a set of required packages to install in the build environment.
        """

    @abc.abstractmethod
    def get_build_environment(self) -> Dict[str, str]:
        """
        Return a dictionary with the environment to use in the build step.

        For consistency, the keys should be defined as SNAPCRAFT_<PLUGIN>_<KEY>.

        This method is called by the PluginHandler during the "build" step.
        """

    @property
    def out_of_source_build(self):
        """Set to True if the plugin performs out-of-source-tree builds.

        In practice, this controls whether the PluginHandler code will
        copy the source code to the build directory before invoking
        the plugin's build commands.
        """
        return False

    @abc.abstractmethod
    def get_build_commands(self) -> List[str]:
        """
        Return a list of commands to run during the build step.

        This method is called by the PluginHandler during the "build" step.
        These commands are run in a single shell instance. This means
        that commands run before do affect the commands that follow.

        snapcraftctl can be used in the script to call out to snapcraft
        specific functionality.
        """



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
from unittest import mock

from testtools.matchers import Equals, MatchesRegex

from snapcraft.plugins.v1._python import _python_finder, errors

from ._basesuite import PythonBaseTestCase


class GetPythonCommandTestCase(PythonBaseTestCase):
    def test_staged(self):
        """get_python_command should support staged python"""

        stage_dir = "stage_dir"
        install_dir = "install_dir"

        # Create staged binary
        self._create_python_binary(stage_dir)

        python_command = _python_finder.get_python_command(
            "test", stage_dir=stage_dir, install_dir=install_dir
        )
        self.assertThat(
            python_command, Equals(os.path.join(stage_dir, "usr", "bin", "pythontest"))
        )

    def test_in_part(self):
        """get_python_command should support in-part python"""

        stage_dir = "stage_dir"
        install_dir = "install_dir"

        # Create installed binary
        self._create_python_binary(install_dir)

        python_command = _python_finder.get_python_command(
            "test", stage_dir=stage_dir, install_dir=install_dir
        )
        self.assertThat(
            python_command,
            Equals(os.path.join(install_dir, "usr", "bin", "pythontest")),
        )

    def test_staged_and_in_part(self):
        """get_python_command should prefer staged python over in-part"""

        stage_dir = "stage_dir"
        install_dir = "install_dir"

        # Create both staged and installed binaries
        self._create_python_binary(stage_dir)
        self._create_python_binary(install_dir)

        python_command = _python_finder.get_python_command(
            "test", stage_dir=stage_dir, install_dir=install_dir
        )
        self.assertThat(
            python_command, Equals(os.path.join(stage_dir, "usr", "bin", "pythontest"))
        )

    def test_missing_raises(self):
        """get_python_command should raise if no python can be found"""

        stage_dir = "stage_dir"
        install_dir = "install_dir"

        raised = self.assertRaises(
            errors.MissingPythonCommandError,
            _python_finder.get_python_command,
            "test",
            stage_dir=stage_dir,
            install_dir=install_dir,
        )
        self.assertThat(
            str(raised),
            Equals("Unable to find pythontest, searched: stage_dir:install_dir"),
        )


class GetPythonHeadersTestCase(PythonBaseTestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("glob.glob")
        self.mock_glob = patcher.start()
        self.addCleanup(patcher.stop)

    def test_staged(self):
        """get_python_headers should support staged headers"""

        stage_dir = "stage_dir"

        # Fake out glob so it looks like the headers are in the staging area
        def _fake_glob(pattern):
            if pattern.startswith(stage_dir):
                return [os.path.join(stage_dir, "usr", "include", "pythontest")]
            return []

        self.mock_glob.side_effect = _fake_glob

        self.assertThat(
            _python_finder.get_python_headers("test", stage_dir=stage_dir),
            Equals(os.path.join(stage_dir, "usr", "include", "pythontest")),
        )

    def test_host(self):
        """get_python_headers should support the hosts's headers"""

        stage_dir = "stage_dir"

        # Fake out glob so it looks like the headers are installed on the host
        def _fake_glob(pattern):
            if pattern.startswith(os.sep):
                return [os.path.join(os.sep, "usr", "include", "pythontest")]
            return []

        self.mock_glob.side_effect = _fake_glob

        self.assertThat(
            _python_finder.get_python_headers("test", stage_dir=stage_dir),
            MatchesRegex(
                re.escape(os.path.join(os.sep, "usr", "include", "pythontest"))
            ),
        )

    def test_staged_and_host(self):
        """get_python_headers should prefer staged headers over host"""

        stage_dir = "stage_dir"

        # Fake out glob so it looks like the headers are in the staging area
        # AND on the host
        def _fake_glob(pattern):
            if pattern.startswith(stage_dir):
                return [os.path.join(stage_dir, "usr", "include", "pythontest")]
            elif pattern.startswith(os.sep):
                return [os.path.join(os.sep, "usr", "include", "pythontest")]
            return []

        self.mock_glob.side_effect = _fake_glob

        self.assertThat(
            _python_finder.get_python_headers("test", stage_dir=stage_dir),
            Equals(os.path.join(stage_dir, "usr", "include", "pythontest")),
        )

    def test_none(self):
        """get_python_headers should return an empty string if no headers"""

        stage_dir = "stage_dir"

        # Fake out glob so it looks like the headers aren't installed anywhere
        def _fake_glob(pattern):
            return []

        self.mock_glob.side_effect = _fake_glob

        self.assertFalse(_python_finder.get_python_headers("test", stage_dir=stage_dir))


class GetPythonHomeTestCase(PythonBaseTestCase):
    def test_staged(self):
        """get_python_home should support staged python"""

        stage_dir = "stage_dir"
        install_dir = "install_dir"

        # Create staged binary
        self._create_python_binary(stage_dir)

        python_home = _python_finder.get_python_home(
            "test", stage_dir=stage_dir, install_dir=install_dir
        )
        self.assertThat(python_home, Equals(os.path.join(stage_dir, "usr")))

    def test_in_part(self):
        """get_python_home should support in-part python"""

        stage_dir = "stage_dir"
        install_dir = "install_dir"

        # Create installed binary
        self._create_python_binary(install_dir)

        python_home = _python_finder.get_python_home(
            "test", stage_dir=stage_dir, install_dir=install_dir
        )
        self.assertThat(python_home, Equals(os.path.join(install_dir, "usr")))

    def test_staged_and_in_part(self):
        """get_python_home should prefer staged python over in-part"""

        stage_dir = "stage_dir"
        install_dir = "install_dir"

        # Create both staged and installed binaries
        self._create_python_binary(stage_dir)
        self._create_python_binary(install_dir)

        python_home = _python_finder.get_python_home(
            "test", stage_dir=stage_dir, install_dir=install_dir
        )
        self.assertThat(python_home, Equals(os.path.join(stage_dir, "usr")))

    def test_missing_raises(self):
        """get_python_home should raise if no python can be found"""

        stage_dir = "stage_dir"
        install_dir = "install_dir"

        raised = self.assertRaises(
            errors.MissingPythonCommandError,
            _python_finder.get_python_home,
            "test",
            stage_dir=stage_dir,
            install_dir=install_dir,
        )
        self.assertThat(
            str(raised),
            Equals("Unable to find pythontest, searched: stage_dir:install_dir"),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from unittest import mock

import fixtures
from testtools import TestCase
from testtools.matchers import Contains

from snapcraft.internal.lifecycle._runner import _install_build_snaps


class TestSnapInstall(TestCase):
    def setUp(self):
        super().setUp()

        self.fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(self.fake_logger)

    @mock.patch("snapcraft.repo.snaps.install_snaps")
    def test_install(self, mock_install_build_snaps):
        _install_build_snaps({"foo/latest/stable", "bar/default/edge"}, set())

        mock_install_build_snaps.assert_called_once_with(
            {"foo/latest/stable", "bar/default/edge"}
        )

    @mock.patch("snapcraft.repo.snaps.install_snaps")
    def test_install_with_content_snap(self, mock_install_build_snaps):
        _install_build_snaps({"foo/latest/stable"}, {"content1/latest/stable"})

        mock_install_build_snaps.assert_has_calls(
            [mock.call({"foo/latest/stable"}), mock.call(["content1/latest/stable"])]
        )

    @mock.patch("snapcraft.internal.common.is_process_container", return_value=True)
    @mock.patch("snapcraft.repo.snaps.install_snaps")
    def test_install_on_docker(self, mock_install_build_snaps, mock_docker_instance):
        _install_build_snaps({"foo/latest/stable", "bar/default/edge"}, set())

        mock_install_build_snaps.assert_not_called()
        self.assertThat(
            self.fake_logger.output,
            Contains(
                "The following snaps are required but not installed as snapcraft "
                "is running inside docker or podman container: "
            ),
        )

    @mock.patch("snapcraft.internal.common.is_process_container", return_value=True)
    @mock.patch("snapcraft.repo.snaps.install_snaps")
    def test_install_with_content_snap_on_docker(
        self, mock_install_build_snaps, mock_docker_instance
    ):
        _install_build_snaps({"foo/latest/stable"}, {"content1/latest/stable"})

        mock_install_build_snaps.assert_not_called()
        self.assertThat(
            self.fake_logger.output,
            Contains(
                "The following snaps are required but not installed as snapcraft "
                "is running inside docker or podman container: "
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from typing import Any

from snapcraft import yaml_utils


class InfoFile(dict):
    """A helper to persist dictionaries as yaml files."""

    def __init__(self, path: str) -> None:
        self._path = path

    def load(self):
        if os.path.exists(self._path):
            with open(self._path) as info_file:
                self.clear()
                self.update(yaml_utils.load(info_file))

    def save(self, **data: Any) -> None:
        dirpath = os.path.dirname(self._path)
        if dirpath:
            os.makedirs(dirpath, exist_ok=True)
        with open(self._path, "w") as info_file:
            data.update(dict(self))
            yaml_utils.dump(data, stream=info_file)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import os
from datetime import datetime
from pathlib import Path
from typing import Set

from snapcraft.internal.deprecations import handle_deprecation_notice
from snapcraft.internal.meta.snap import Snap

from ._project_info import ProjectInfo  # noqa: F401
from ._project_options import ProjectOptions


class Project(ProjectOptions):
    """All details around building a project concerning the build environment
    and the snap being built."""

    def __init__(
        self,
        *,
        target_deb_arch=None,
        debug=False,
        snapcraft_yaml_file_path=None,
        work_dir: str = None,
        is_managed_host: bool = False
    ) -> None:

        project_dir = os.getcwd()
        if is_managed_host:
            work_dir = os.path.expanduser("~")
        else:
            work_dir = project_dir

        super().__init__(target_deb_arch, debug, work_dir=work_dir)

        # This here check is mostly for backwards compatibility with the
        # rest of the code base.
        if snapcraft_yaml_file_path is None:
            self.info: ProjectInfo = None  # type: ignore

        else:
            self.info = ProjectInfo(snapcraft_yaml_file_path=snapcraft_yaml_file_path)

        self._is_managed_host = is_managed_host
        self._project_dir = project_dir
        self._work_dir = work_dir

        self.local_plugins_dir = self._get_local_plugins_dir()
        self._start_time = datetime.utcnow()

        # XXX: (Re)set by Config because it mangles source data.
        # Ideally everywhere wold converge to operating on snap_meta, and ww
        # would only need to initialize it once (properly).
        self._snap_meta = Snap()

    def _get_build_base(self) -> str:
        """
        Return name for type base or the base otherwise build-base is set
        """

        # In case snap_meta is not yet populated, lookup base in info.
        if self.info:
            if self.info.build_base:
                return self.info.build_base
            if self.info.base:
                return self.info.base

        return self._snap_meta.get_build_base()

    def _get_project_directory_hash(self) -> str:
        m = hashlib.sha1()
        m.update(self._project_dir.encode())
        return m.hexdigest()[:6]

    def _get_content_snaps(self) -> Set[str]:
        """Return the set of content snaps from snap_meta."""
        return set(
            [
                x.provider
                for x in self._snap_meta.get_content_plugs()
                if x.provider is not None
            ]
        )

    def _get_provider_content_dirs(self) -> Set[str]:
        """Return the set installed provider content directories."""
        return self._snap_meta.get_provider_content_directories()

    def _get_snapcraft_assets_dir(self) -> str:
        # Many test cases don't set the yaml file path and assume the default dir
        if not self.info:
            return os.path.join(self._project_dir, "snap")

        if self.info.snapcraft_yaml_file_path.endswith(
            os.path.join("build-aux", "snap", "snapcraft.yaml")
        ):
            return os.path.join(self._project_dir, "build-aux", "snap")
        else:
            return os.path.join(self._project_dir, "snap")

    def _get_keys_path(self) -> Path:
        # Directory containing <KEY_ID>.asc keys for use with
        # package-repositories, relative to 'snap' assets.
        return Path(self._get_snapcraft_assets_dir(), "keys")

    def _get_local_plugins_dir(self) -> str:
        deprecated_plugins_dir = os.path.join(self._parts_dir, "plugins")
        if os.path.exists(deprecated_plugins_dir):
            handle_deprecation_notice("dn2")
            return deprecated_plugins_dir
        else:
            assets_dir = self._get_snapcraft_assets_dir()
            return os.path.join(assets_dir, "plugins")

    def _get_global_state_file_path(self) -> str:
        if self._is_managed_host:
            state_file_path = os.path.join(self._work_dir, "state")
        else:
            state_file_path = os.path.join(self._parts_dir, ".snapcraft_global_state")

        return state_file_path

    def _get_stage_packages_target_arch(self) -> str:
        """Get architecture for staging packages.

        Prior to core20, staging packages has broken behavior in that it will
        stage native architecture packages by default.

        :return: The appropriate default architecture to stage.
        """
        if self._get_build_base() in ["core16", "core18"]:
            return self.deb_arch
        else:
            return self.target_arch

    def _get_start_time(self) -> datetime:
        """Returns the timestamp for when a snapcraft project was loaded."""
        return self._start_time

    def is_legacy_project(self) -> bool:
        """Determine if project is legacy."""
        # Not in an actual project.
        if self.info is None:
            return False
        # Building a base is not legacy.
        elif self.info.type == "base":
            return False
        elif self.info.build_base is not None:
            return False
        # A self with no base set is legacy.
        elif self.info.base is None:
            return True
        else:
            return False
